from __future__ import annotations
from typing import Optional, TYPE_CHECKING, TypeVar, Generic, Any, Iterator, Type
from .proxy import Proxy
from .proxylist import ProxyList, ProxySublist
from .property import Property
from .refs import Reference, RefList

if TYPE_CHECKING:
    from ..client import Client
    from .property import RefList
    from .registry import Registry
    from .schema import Schema

ProxyClass = TypeVar('ProxyClass', bound=Proxy)

class ProxySynclist:
    """A container for the proxies that need to be sync'd after
       an operation.
    """
    
    def __init__(self, l: list[Proxy] = []):
        self.tosync = list()
        for a in l:
            self.add(a)
    
    def add(self, ref: Proxy|ProxyList):
        """Add a proxy or the elements of a proxy list to the synclist.

            Any other types are ignored.
        """
        if isinstance(ref, ProxyList):
            for p in ref:
                self.add(p)
        elif isinstance(ref, Proxy):
            self.tosync.append(ref)
            
    def trigger_properties(self, schema: Schema):
        for p in schema.properties.values():
            if isinstance(p, Reference) and p.trigger_sync:
                yield p

    def on_create(self, proxy_type: Type[ProxyClass], properties: dict[str, Any]):
        for p in self.trigger_properties(proxy_type.proxy_schema):
            self.add(properties.get(p.name))

    def on_create_proxy(self, proxy: ProxyClass):
        for p in self.trigger_properties(proxy.proxy_schema):
            self.add(p.get(proxy))

    def on_delete(self, proxy: Proxy):
        for p in proxy.proxy_schema.properties.values():
            self.add( getattr(proxy, p.name) )

    def on_update(self, proxy: Proxy, prop: Property, newref: Proxy):
        if isinstance(prop, Reference) and prop.trigger_sync:
            oldref = getattr(proxy, prop.name)
            self.add(oldref)
            self.add(newref)

    def sync(self):
        for prx in self.tosync:
            prx.proxy_sync()
        self.tosync.clear()





from __future__ import annotations

import pytest

from stelar.client import TaskSpec


@pytest.mark.skip()
def test_exec_word_count_manual(test_client):
    """
    Execute the word count task manually.
    """

    c = test_client
    d1 = c.datasets["shakespeare_novels"]

    # Create a process
    p1 = c.processes.get("word-count-tests")

    api_json = {
        "process_id": str(p1.id),
        # name: Optional, goes into tags
        "name": "Word Count on RJ",
        # Package name or id of the tool
        # If image is not provided, take the 'latest' (most recent) image of the tool
        # from the embedded repository
        # If image is provided use the image
        ### "tool": ""
        "image": "petroud/stelar-word-count:latest",
        #
        "inputs": {"text_to_count": ["d1::owned"]},
        "datasets": {
            "d1": str(d1.id),
            "d2": {
                "name": "the-results-of-word-count",
                "owner_org": "stelar-klms",
                "notes": "The result of map reduce word count",
                "tags": ["wordcount", "map-reduce"],
            },
        },
        # Use anything, not used
        "parameters": {"key": "value", "key1": 1},
        "outputs": {
            "word_count_file": {
                "url": "s3://test-bucket/word-count.txt",
                "dataset": "d2",
                "resource": {"name": "RJ Word Count 2", "relation": "owned"},
            }
        },
        "secrets": {"hello": "world"},
    }

    # Define the input and output files
    resp = c.POST("v2/task", **api_json)

    assert resp.status_code in range(
        200, 300
    ), f"Error: {resp.status_code} - {resp.text}"
    assert (
        resp.json()["status"] == "success"
    ), f"Error: {resp.status_code} - {resp.text}"


def test_task_spec(testcli):
    d1 = testcli.datasets["dataset1"]

    t = TaskSpec(name="Word Count on RJ")
    t.d("d1", d1)
    t.d(
        "d2",
        name="the-results-of-word-count",
        owner_org="stelar-klms",
        notes="The result of map reduce word count",
        tags=["wordcount", "map-reduce"],
    )
    t.i(text_to_count="d1::owned", profile=["d1::profile"])
    t.o(
        word_count_file={
            "url": "s3://test-bucket/word-count.txt",
            "dataset": "d2",
            "resource": {"name": "RJ Word Count 2", "relation": "owned"},
        }
    )
    t.p(key="value", key1=1)

    spec = t.spec()
    target = {
        "name": "Word Count on RJ",
        "datasets": {
            "d1": str(d1.id),
            "d2": {
                "name": "the-results-of-word-count",
                "owner_org": "stelar-klms",
                "notes": "The result of map reduce word count",
                "tags": ["wordcount", "map-reduce"],
            },
        },
        "inputs": {"text_to_count": ["d1::owned"], "profile": ["d1::profile"]},
        "outputs": {
            "word_count_file": {
                "url": "s3://test-bucket/word-count.txt",
                "dataset": "d2",
                "resource": {"name": "RJ Word Count 2", "relation": "owned"},
            }
        },
        "parameters": {
            "key": "value",
            "key1": 1,
        },
    }

    assert set(spec.keys()) == set(target.keys())
    for key in spec:
        assert spec[key] == target[key]


def test_task_spec_empty():
    t = TaskSpec()
    spec = t.spec()
    target = {
        "name": "task",
        "datasets": {},
        "inputs": {},
        "outputs": {},
        "parameters": {},
    }

    assert set(spec.keys()) == set(target.keys())
    for key in spec:
        assert spec[key] == target[key]


def test_task_spec_init(testcli):
    """
    Test the initialization of TaskSpec.
    """

    assert TaskSpec().tool is None
    assert TaskSpec().name == "task"
    assert TaskSpec().image is None

    assert TaskSpec("test-tool").tool == "test-tool"

    tool = testcli.tools["simple_tool"]
    assert TaskSpec(tool).tool == "simple_tool"
    assert TaskSpec(tool.id).tool == str(tool.id)
    assert TaskSpec(tool.name).tool == tool.name

    assert TaskSpec(tool).image is None
    assert TaskSpec(tool, image="latest").image == "latest"
    assert (
        TaskSpec(image="petroud/stelar-word-count:latest").image
        == "petroud/stelar-word-count:latest"
    )
    assert TaskSpec(image="petroud/stelar-word-count:latest").tool is None


def test_task_spec_dataset(testcli):
    """
    Test the dataset method of TaskSpec.
    """

    d1 = testcli.datasets["dataset1"]
    t = TaskSpec().d("d1", d1)
    assert t.datasets == {"d1": str(d1.id)}

    t = TaskSpec().d("d2", name="test-dataset", owner_org="stelar-klms")
    assert t.datasets == {"d2": {"name": "test-dataset", "owner_org": "stelar-klms"}}

    with pytest.raises(ValueError):
        t = TaskSpec().d("d3", d1, name="test-dataset", owner_org="stelar-klms")

    deforg = testcli.datasets.default_organization
    assert TaskSpec().d("d", name="foo", owner_org=deforg).datasets["d"][
        "owner_org"
    ] == str(deforg.id)

    assert TaskSpec().d("d", name="foo", owner_org=deforg.id).datasets["d"][
        "owner_org"
    ] == str(deforg.id)


def test_task_spec_input(testcli):
    """
    Test the input method of TaskSpec.
    """

    d = testcli.datasets["shakespeare_novels"]
    r = d.resources[0]
    rid = str(r.id)

    assert TaskSpec().i(inp=r).spec()["inputs"]["inp"] == [rid]
    assert TaskSpec().i(inp=rid).spec()["inputs"]["inp"] == [rid]

    assert TaskSpec().i(inp=[r]).spec()["inputs"]["inp"] == [rid]
    assert TaskSpec().i(inp=[rid]).spec()["inputs"]["inp"] == [rid]

    assert TaskSpec().i(inp=[r, r]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[rid, rid]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[r, rid]).spec()["inputs"]["inp"] == [rid, rid]
    assert TaskSpec().i(inp=[rid, r]).spec()["inputs"]["inp"] == [rid, rid]


def test_task_spec_output(testcli):
    """
    Test the output method of TaskSpec.
    """

    d = testcli.datasets["shakespeare_novels"]
    r = d.resources[0]
    rid = str(r.id)

    outspec = {"url": r.url}
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
    }

    outspec = {"url": r.url, "resource": r}
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
        "resource": str(r.id),
    }

    outspec = {
        "url": r.url,
        "dataset": str(d.id),
        "resource": {"name": r.name, "relation": "owned"},
    }
    assert TaskSpec().o(out=outspec).spec()["outputs"]["out"] == {
        "url": r.url,
        "dataset": str(d.id),
        "resource": {"name": r.name, "relation": "owned"},
    }


#
# TODO: Add more tests for task execution
#
def test_create_task(testcli):
    """
    Test the creation of a task.
    """

    d = testcli.datasets["shakespeare_novels"]
    proc = testcli.processes["word-count-tests"]



from urllib.parse import urljoin

from requests.exceptions import HTTPError

from .base import BaseAPI
from .endpoints import APIEndpointsV1
from .model import DuplicateEntryError, MissingParametersError, STELARUnknownError
from .policy import Policy


class AdminAPI(BaseAPI):
    """
    Represents a class that handles administrative API actions, including policy and user management.
    """

    def create_policy(self, policy: Policy):
        """
        Create a policy by sending its content to the specified API endpoint.

        This method sends a POST request to the API with the policy content in YAML format.
        If the policy is successfully created, the method updates the input `Policy` object with
        information about the policy details.

        Parameters
        ----------
        policy : Policy
            The `Policy` object containing the policy content to be created.

        Returns
        -------
        Policy
            A `Policy` object representing the policy details.

        Raises
        ------
        MissingParametersError
            If the API responds with a 400 status code, indicating a bad request due
            to missing or incorrect parameters.
        DuplicateEntryError
            If the API responds with a 409 status code, indicating that the dataset
            already exists.
        STELARUnknownError
            If the API responds with an unexpected error, an `STELARUnknownError`
            is raised with the error details.

        Examples
        --------
        policy = Policy(policy_content="your policy content in YAML format")
        admin.create_policy(policy)
        display(policy)

        See Also
        --------
        APIEndpointsV1.POST_POLICY : The API endpoint for creating policies.
        """

        if not policy:
            return None
        try:
            yaml_headers = {"Content-Type": "application/x-yaml"}
            policy_response = self.api_request(
                "POST",
                APIEndpointsV1.POST_POLICY,
                headers=yaml_headers,
                data=policy.policy_content,
            )
            if policy_response.status_code == 200:
                policy_repr = self.api_request(
                    "GET", urljoin(APIEndpointsV1.GET_POLICY_REPRESENATION, "active")
                )
                policy_info = self.api_request(
                    "GET", urljoin(APIEndpointsV1.GET_POLICY_INFO, "active")
                )
                if policy_info.status_code == 200 and policy_repr.status_code == 200:
                    yaml_content = policy_repr.content
                    if yaml_content.startswith(b"b'"):
                        yaml_content = yaml_content[2:-1]

                    formatted_yaml_string = yaml_content.decode("unicode_escape")
                    pjson = policy_info.json()["result"]["policy"]
                    pjson["policy_content"] = formatted_yaml_string
                    # print(yaml.dump(policy_response.json()['result']['policy']))

                    policy.update_from_dict(pjson)
                    policy.reset_dirty()
        except HTTPError as he:
            if he.response.status_code == 400:
                raise MissingParametersError("Bad Request")
            elif he.response.status_code == 409:
                raise DuplicateEntryError("Dataset Already Exists")
            else:
                raise STELARUnknownError(f"Unknown Error: {he}")

    def get_policy_info(self, filter: str):
        """
        Retrieve detailed information about a specific policy based on a filter.

        Parameters
        ----------
        filter : str
            The filter string to identify the policy.

        Returns
        -------
        Policy
            A `Policy` object representing the policy details.

        Raises
        ------
        MissingParametersError
            If the API responds with a 400 status code.
        DuplicateEntryError
            If the API responds with a 409 status code.
        STELARUnknownError
            If an unexpected error occurs.

        Examples
        --------
        policy = admin.get_policy_info("policy-filter")
        display(policy)
        """
        if not filter:
            return None
        try:
            policy_repr = self.api_request(
                "GET", urljoin(APIEndpointsV1.GET_POLICY_REPRESENATION, filter)
            )
            policy_info = self.api_request(
                "GET", urljoin(APIEndpointsV1.GET_POLICY_INFO, filter)
            )
            if policy_info.status_code == 200 and policy_repr.status_code == 200:
                yaml_content = policy_repr.content
                if yaml_content.startswith(b"b'"):
                    yaml_content = yaml_content[2:-1]

                formatted_yaml_string = yaml_content.decode("unicode_escape")
                pjson = policy_info.json()["result"]["policy"]
                pjson["policy_content"] = formatted_yaml_string
                policy_obj = Policy.from_dict(pjson)
                return policy_obj
        except HTTPError as he:
            if he.response.status_code == 400:
                raise MissingParametersError("Bad Request")
            elif he.response.status_code == 409:
                raise DuplicateEntryError("Dataset Already Exists")
            else:
                raise STELARUnknownError(f"Unknown Error: {he}")

    def get_policy_representation(self, filter: str):
        """
        Retrieve the YAML representation of a specific policy based on a filter.

        Parameters
        ----------
        filter : str
            The filter string to identify the policy.

        Returns
        -------
        None

        Raises
        ------
        MissingParametersError
            If the API responds with a 400 status code.
        DuplicateEntryError
            If the API responds with a 409 status code.
        STELARUnknownError
            If an unexpected error occurs.

        Examples
        --------
        admin.get_policy_representation("policy-filter")
        """
        if not filter:
            return None
        try:
            policy_repr = self.api_request(
                "GET", urljoin(APIEndpointsV1.GET_POLICY_REPRESENATION, filter)
            )
            if policy_repr.status_code == 200:
                yaml_content = policy_repr.content
                if yaml_content.startswith(b"b'"):
                    yaml_content = yaml_content[2:-1]

                formatted_yaml_string = yaml_content.decode("unicode_escape")
                print(formatted_yaml_string)
        except HTTPError as he:
            if he.response.status_code == 400:
                raise MissingParametersError("Bad Request")
            elif he.response.status_code == 409:
                raise DuplicateEntryError("Dataset Already Exists")
            else:
                raise STELARUnknownError(f"Unknown Error: {he}")

    def get_policy_list(self):
        """
        Retrieve a list of all available policies and present them as tables.

        Returns
        -------
        Policy List
            A List of dictionaries containing policy information(policy_uuid, policy_familiar_name)

        Raises
        ------
        MissingParametersError
            If the API responds with a 400 status code.
        DuplicateEntryError
            If the API responds with a 409 status code.
        STELARUnknownError
            If an unexpected error occurs.

        Examples
        --------
        admin.get_policy_list()
        """
        try:
            policy_response = self.api_request("GET", APIEndpointsV1.GET_POLICY_LIST)
            if policy_response.status_code == 200:
                pjson = policy_response.json()["result"]["policies"]
                policy_list = [json for json in pjson]
                return policy_list
                # Policy.present_dictionaries_as_tables(policy_list)
        except HTTPError as he:
            if he.response.status_code == 400:
                raise MissingParametersError("Bad Request")
            elif he.response.status_code == 409:
                raise DuplicateEntryError("Dataset Already Exists")
            else:
                raise STELARUnknownError(f"Unknown Error: {he}")



from IPython.core.display import HTML
from IPython.display import display


class Policy:
    """
    A class representing a STELAR policy with metadata and additional details.

    This class is designed to hold metadata for a STELAR policy entity, making it easier
    to manage and manipulate policy information within the client application runtime.
    """

    def __init__(self, policy_familiar_name: str, policy_content: str | dict) -> None:
        """
        Initialize a Policy instance with essential fields.

        Args:
            familiar_name (str): The name of the policy to be applied.
            policy_content (str | dict): The yaml file describing the
                        roles and permissions to be applied
        """
        self._data = {
            "policy_familiar_name": policy_familiar_name,
        }
        self._original_data = self._data.copy()  # Copy for dirty tracking
        self._dirty_fields = set()  # Track which fields have been modified

        # Additional fields
        self.active = None
        self.created_at = None
        self.policy_uuid = None
        self.user_id = None

        self.policy_content = self._parse_policy_content(policy_content)

    def _parse_policy_content(self, content: str | dict) -> str:
        """
        Return the raw YAML content (as string or file content).

        Args:
            content (str | dict): YAML string, file path, or dictionary.

        Returns:
            str: Raw YAML content as a string.
        """
        if isinstance(content, dict):
            raise ValueError(
                "Expected YAML content as string or file path, not a dictionary."
            )

        try:
            if isinstance(content, str):
                if "\n" in content or ":" in content:  # Likely a YAML string
                    return content  # Return YAML string as is
                else:  # Assume it's a file path
                    with open(content, "r") as file:
                        return file.read()  # Return raw file content as string
            else:
                raise ValueError("Content must be a YAML string or a file path.")

        except Exception as e:
            raise ValueError(f"Invalid YAML content: {e}")

    def __getattr__(self, name):
        """Handle attribute access dynamically."""
        if name in self._data:
            return self._data[name]
        raise AttributeError(f"'Resource' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        """Handle attribute assignment dynamically."""
        if (
            name in {"_data", "_original_data", "_dirty_fields"}
            or name not in self._data
        ):
            super().__setattr__(name, value)
        else:
            # Only mark as dirty if the value actually changes
            if self._data[name] != value:
                self._data[name] = value
                self._dirty_fields.add(name)

    def is_dirty(self):
        """Check if any field is dirty."""
        return bool(self._dirty_fields)

    def changes(self):
        """Return the fields that have been modified and their changes."""
        return {
            key: (self._original_data[key], self._data[key])
            for key in self._dirty_fields
        }

    def reset_dirty(self):
        """Reset the dirty state and save the current data as original."""
        self._original_data = self._data.copy()
        self._dirty_fields.clear()

    @classmethod
    def from_dict(cls, data: dict):
        """
        A class method to construct a Policy instance from a dictionary.

        Args:
            data (dict): The dictionary containing policy metadata.

        Returns:
            Policy: A fully constructed Policy instance.
        """
        instance = cls(
            policy_familiar_name=data.get("policy_familiar_name"),
            policy_content=data.get("policy_content"),
        )
        instance._populate_additional_fields(data)
        return instance

    def _populate_additional_fields(self, data: dict):
        """
        Populate additional fields of the Policy object.

        Args:
            data (dict): The dictionary containing policy metadata.
        """
        self.active = data.get("active")
        self.created_at = data.get("created_at")
        self.policy_uuid = data.get("policy_uuid")
        self.user_id = data.get("user_id")
        self.policy_content = data.get("policy_content")

        return self

    def to_dict(self):
        """
        Convert the Resource instance into a dictionary.

        Returns:
            dict: A dictionary representation of the Resource instance.
        """
        policy_dict = {
            "policy_familiar_name": self.policy_familiar_name,
            "yaml_content": self.policy_content,
        }
        return policy_dict

    def update_from_dict(self, data: dict):
        """
        Updates the current Resource instance with new data from a dictionary.

        Args:
            data (dict): A dictionary containing new data for the Resource.
        """
        for key in ["policy_familiar_name"]:
            if key in data:
                setattr(self, key, data[key])

        # Update additional fields
        self._populate_additional_fields(data)

    def _repr_html_(self):
        """
        Provide an HTML representation of the Policy instance for Jupyter display,
        with an extra cell named 'Policy Representation' displaying the policy content,
        and a watermark image behind the table.
        """
        # Format policy content as YAML (if present)
        policy_content_value = (
            f"<pre>{self.policy_content}</pre>"
            if self.policy_content
            else "No Policy Content"
        )

        # Build the table rows for the policy attributes
        rows = f"""
            <tr style="background-color: rgba(255, 255, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">active:</td>
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{self.active or 'N/A'}</td>
            </tr>
            <tr style="background-color: rgba(230, 179, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">created_at:</td>
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{self.created_at or 'N/A'}</td>
            </tr>
            <tr style="background-color: rgba(255, 255, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">policy_familiar_name:</td>
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{self.policy_familiar_name}</td>
            </tr>
            <tr style="background-color: rgba(230, 179, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">policy_uuid:</td>
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{self.policy_uuid or 'N/A'}</td>
            </tr>
            <tr style="background-color: rgba(255, 255, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">user_id:</td>
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{self.user_id or 'N/A'}</td>
            </tr>
            <tr style="background-color: rgba(230, 179, 255, 0.8); color: black;">
                <td style="text-align: center; padding: 5px; border: 1px solid #ddd; font-weight: bold; width: 50%;">Policy Representation:</td>
                <td style="text-align: left; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{policy_content_value}</td>
            </tr>
        """

        # Define the watermark image URL (Replace with your image path)
        watermark_image_url = "logo.png"

        # Build the complete HTML with "Policy Summary" centered above the table and watermark
        html = f"""
        <div style="position: relative; width: 100%; height: auto; display: flex; justify-content: flex-start; align-items: flex-start;">
            <table border="1" style="
                border-collapse: collapse;
                width: 50%;
                margin: 0;
                color: black;
                background-image: url('{watermark_image_url}');
                background-size: 60%;
                background-position: center;
                background-repeat: no-repeat;">
                <thead>
                    <tr>
                        <th colspan="2" style="text-align: center; padding: 10px; font-size: 18px; font-weight: bold; background-color: rgba(242, 242, 242, 0.8);">
                            Policy Summary
                        </th>
                    </tr>
                </thead>
                <tbody>
                    {rows}
                </tbody>
            </table>
        </div>
        """
        return HTML(html)._repr_html_()

    def present_dictionaries_as_tables(dicts_list):
        """
        Display each dictionary as an individual HTML table with attributes on the left,
        values on the right, bolded attribute text, and black text for visibility.
        The values column is positioned closer to the split line.

        Parameters:
            dicts_list (list of dict): A list of dictionaries to present.
        """
        # Generate HTML tables with the desired style
        html_content = ""
        for idx, data in enumerate(dicts_list):
            # Build table rows with specified styles
            rows = "\n".join(
                f"""
                <tr style="background-color: {('rgba(255, 255, 255, 0.8)' if i % 2 == 0 else 'rgba(230, 179, 255, 0.8)')}; color: black;">
                    <td style="text-align: left; padding: 5px; border: 1px solid #ddd; font-weight: bold; color: black; width: 50%;">{key}:</td>
                    <td style="text-align: left; padding: 5px; border: 1px solid #ddd; color: black; width: 50%;">{value}</td>
                </tr>
                """
                for i, (key, value) in enumerate(data.items())
            )

            # Define the watermark image URL (Replace with your image path)
            watermark_image_url = "logo.png"

            # Build table HTML
            table_html = f"""
            <div style="position: relative; width: 100%; height: auto; display: flex; justify-content: flex-start; align-items: flex-start; margin-bottom: 20px;">
                <table border="1" style="
                    border-collapse: collapse;
                    width: 50%;
                    margin: 0;
                    color: black;
                    background-image: url('{watermark_image_url}');
                    background-size: 20%;
                    background-position: center;
                    background-repeat: no-repeat;">
                    <thead>
                        <tr>
                            <th colspan="2" style="text-align: center; padding: 10px; font-size: 18px; font-weight: bold; background-color: rgba(242, 242, 242, 0.8);">
                                Policy Summary
                            </th>
                        </tr>
                    </thead>
                    <tbody>
                        {rows}
                    </tbody>
                </table>
            </div>
            """
            html_content += table_html

        # Render the content
        display(HTML(html_content))



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, Type, TypeVar
from uuid import UUID

import pandas as pd

from .proxy import Proxy

if TYPE_CHECKING:
    from ..client import Client
    from .refs import RefList


ProxyClass = TypeVar("ProxyClass", bound=Proxy)


class ShortenedUUID(UUID):
    def __str__(self):
        return f"{super().__str__()[:5]}..."


def simplified(value, property):
    from .property import Id
    from .refs import Reference

    match property:
        case Id():
            return ShortenedUUID(value.hex)
        case Reference(proxy_type=pt):
            if pt.proxy_schema.name_id:
                return value.name
            elif isinstance(value, Proxy):
                return ShortenedUUID(value.proxy_id.hex)
            elif isinstance(value, ProxyList):
                return f"{pt.proxy_schema.class_name}[{len(value)}]"
            else:
                return value
        case _:
            return value


class ProxyList(Generic[ProxyClass]):
    """
    Base class for "dynamic lists" of proxies.

    The instances maintain lists of IDs and translate list operations
    as if the list actually contained proxy objects.
    """

    def __init__(self, client: Client, proxy_type: Type[ProxyClass]):
        self.client = client
        self.proxy_type = proxy_type
        self.registry = client.registry_for(proxy_type)

    @property
    def coll(self) -> list[UUID]:
        raise NotImplementedError

    def resolve_proxy(self, item):
        """This is the main routine that transforms elements of
        the list to proxies.
        """
        return self.registry.fetch_proxy(item)

    def __iter__(self):
        for item in self.coll:
            yield self.resolve_proxy(item)

    def __len__(self):
        return len(self.coll)

    def __getitem__(self, item):
        if isinstance(item, slice):
            raise ValueError("Slices are not supported yet")
        return self.resolve_proxy(self.coll[item])

    def __repr__(self):
        return f"{self.proxy_type.__name__}{repr(self.coll)}"

    def __eq__(self, other):
        try:
            return len(self) == len(other) and (
                all(p == q for p, q in zip(self, other))
            )
        except Exception:
            return False

    @property
    def ids(self):
        return list(self.coll)

    def to_df(self, *additional_fields, fields=None, simplify=True):
        """Generate a pandas dataframe for the list of proxy entities.

        The dataframe is generated by fetching and tabulating a
        subset of fields, for each entity in the list.



        """
        schema = self.proxy_type.proxy_schema

        if fields is None:
            fields = schema.short_list(set(additional_fields))
        elif fields in (True, "all", "ALL"):
            fields = list(schema.all_fields) + list(additional_fields)
        else:
            fields = list(fields) + list(additional_fields)
        data = {field: list() for field in fields}

        for proxy in self:
            for field in fields:
                if simplify:
                    property = schema.all_fields[field]
                    data[field].append(
                        simplified(getattr(proxy, field, pd.NA), property)
                    )
                else:
                    data[field].append(getattr(proxy, field, pd.NA))
        return pd.DataFrame(data=data)

    @property
    def df(self):
        """Return a dataframe over the default fields."""
        return self.to_df()

    @property
    def DF(self):
        """Return a dataframe over all fields."""
        return self.to_df(fields=True, simplify=False)


class ProxyVec(ProxyList):
    """A list of IDs appearing as proxies.

    The underlying data is a list of UUIDs. At each element access, the
    correpsonding element is fetched from the registry.
    """

    def __init__(
        self, client: Client, proxy_type: Type[ProxyClass], members: list[UUID]
    ):
        """Initialize the vector with the client, the proxy type, and the list of IDs.

        Arguments:
            client: the client object
            proxy_type: the type of the proxies
            members: the list of UUIDs
        """
        super().__init__(client, proxy_type)
        self.members = members

    @property
    def coll(self):
        return self.members


class ProxySublist(ProxyList):
    """A proxy class that translates collection operations to operations
    on an entity sub-collection.
    """

    def __init__(self, property: RefList, owner: Proxy):
        super().__init__(owner.proxy_registry.catalog, property.proxy_type)
        self.property = property
        self.owner = owner

    def __delitem__(self, key):
        raise NotImplementedError(
            f"delitem  {self.property.owner.__name__}.{self.property.name}"
        )

    def __iadd__(self, **kwargs):
        raise NotImplementedError(
            f"iadd {self.property.owner.__name__}.{self.property.name}"
        )

    @property
    def coll(self):
        return self.property.get(self.owner)



from .base import BaseAPI
from .package import PackageCursor
from .tasks import TaskCursor
from .workflows import ProcessCursor, ToolCursor, Workflow


class WorkflowsAPI(BaseAPI):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.processes = ProcessCursor(self)
        self.tasks = TaskCursor(self)
        self.workflows = PackageCursor(self, Workflow)
        self.tools = ToolCursor(self)

        self.load_local_task_sigs()

    def load_local_task_sigs(self):
        """Load local task signatures from the local filesystem."""
        import csv
        from pathlib import Path

        lts_filepath = Path.home() / ".stelar_lts.csv"
        if lts_filepath.exists():
            with open(lts_filepath, newline="") as f:
                lts_reader = csv.reader(f, delimiter=",")
                self.local_task_sigs = {row[0]: row[1] for row in lts_reader}
        else:
            self.local_task_sigs = {}

    def commit_local_task_sig(self, task_id: str, sig: str):
        """Commit a local task signature to the local filesystem."""
        import csv
        from pathlib import Path

        lts_filepath = Path.home() / ".stelar_lts.csv"
        with open(lts_filepath, "a", newline="") as f:
            lts_writer = csv.writer(f, delimiter=",")
            lts_writer.writerow([task_id, sig])
        self.local_task_sigs[task_id] = sig



"""Basic support for tasks in the client.
"""
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Generator
from uuid import UUID

from .api_call import api_call
from .dataset import Dataset
from .generic import GenericCursor, GenericProxy
from .group import Organization
from .proxy import DateField, Id, Property, Reference, StrField, derived_property
from .proxy.property import DictProperty, ListProperty
from .resource import Resource
from .utils import client_for

if TYPE_CHECKING:
    # Type definitions for
    from .workflows import Tool

    ToolSpec = Tool | UUID | str | None
    ImageSpec = str | None
    ExistingDataset = Dataset | UUID | str


class Task(GenericProxy):
    id = Id(entity_name="id")
    start_date = Property(validator=DateField)
    end_date = Property(validator=DateField(nullable=True))

    exec_state = Property(validator=StrField)

    creator = Property(validator=StrField, entity_name="creator")
    process = Reference("Process", entity_name="process_id", trigger_sync=False)

    messages = Property(validator=StrField, updatable=False, optional=True)

    # TODO: parameters are dict[str, json]
    parameters = DictProperty(str, str, updatable=False, optional=True)
    metrics = DictProperty(str, str, updatable=False, optional=True)

    inputs = DictProperty(str, list[str], updatable=False, optional=True)
    outputs = ListProperty(dict, updatable=False, optional=True)

    tool = Property(validator=StrField, optional=True)
    image = Property(validator=StrField, optional=True)

    # TODO: return the tool proxy
    # via a method

    tags = DictProperty(str, str, updatable=False, optional=True)

    @derived_property
    def is_external(self, entity) -> bool:
        """Check if this task is local (i.e., not using a remote tool)."""
        return "image" not in entity

    @property
    def signature(self) -> str:
        """Return a string signature for this task."""
        return client_for(self).local_task_sigs.get(str(self.id))

    @cached_property
    def job_input(self, signature=None) -> dict:
        """Return the job input for this local task.

        The job input is a dictionary that contains the inputs and parameters
        of the task, formatted for use in a job execution context.
        """
        if not self.is_external:
            raise ValueError("Cannot get job input for non-external tasks")

        if signature is None:
            signature = self.signature
            if signature is None:
                raise ValueError(
                    "No signature, cannot get job input for a task without a signature"
                )

        return api_call(self).task_job_input(task_id=str(self.id), signature=signature)

    def exit_job(
        self,
        *,
        metrics={},
        output={},
        message="The task is finished",
        status="succeeded",
        signature=None,
    ):
        """Post the job output for this external task.

        The job output contains the outputs and metrics of the task,
        the final message and a task status.

        Args:
            metrics (dict): A dictionary of metrics to post.
            output (dict): A dictionary of outputs to post.
            message (str): A message to post with the job output.
            status (str): The status of the task, e.g., "succeeded" or "failed".
        Raises:
            ValueError: If the task is not external or if the POST request fails.

        """
        if not self.is_external:
            raise ValueError("Cannot post job output for non-external tasks")
        if signature is None:
            signature = self.signature
            if signature is None:
                raise ValueError(
                    "No signature, cannot post job output for a task without a signature"
                )

        output_spec = {
            "metrics": metrics,
            "output": output,
            "message": message,
            "status": status,
        }

        api_call(self).task_post_job_output(str(self.id), signature, output_spec)
        self.proxy_sync()

    def jobs(self):
        """Return a dictionary with info about Kubernetes jobs for this task."""
        response = client_for(self).GET("v2/task", self.id, "jobs")
        response.raise_for_status()
        return response.json()["result"]

    def logs(self):
        """Return the logs for this task."""
        response = client_for(self).GET("v2/task", self.id, "logs")
        response.raise_for_status()
        return response.json()["result"]

    def printlog(self, *, file=None, flush=False):
        for job, log in self.logs().items():
            print(job, ":", file=file, flush=flush)
            print(log, file=file, flush=flush)


class TaskSpec:
    """A task spec is used to create tasks.

    Contrary to other entities, Task creation is more complicated than simply
    providing values for some scalar attributes. Tasks, once created are
    not really editable; therefore, all task fields must be specified
    at task creation time.

    This class helps by collecting arguments needed to initialize a new Task.
    """

    def __init__(
        self, tool: ToolSpec = None, *, image: ImageSpec = None, name: str = "task"
    ):
        from .workflows import Tool

        if isinstance(tool, Tool):
            self.tool = str(tool.name)
        elif isinstance(tool, UUID):
            self.tool = str(tool)
        elif isinstance(tool, str | None):
            self.tool = tool
        else:
            raise TypeError("Expected a Tool proxy or a string, or None")

        if isinstance(image, str | None):
            self.image = image
        else:
            raise TypeError("Expected a string as image")

        if not isinstance(name, str):
            raise TypeError("Expected a string as name")

        self.name = name
        self.datasets = {}
        self.inputs = {}
        self.outputs = {}
        self.parameters = {}

    def is_external(self) -> bool:
        """Check if this task spec is local (i.e., not using a remote tool)."""
        return self.tool is None and self.image is None

    def d(
        self, alias: str, dset: Dataset | UUID | str | None = None, **dspec
    ) -> TaskSpec:
        """Add a dataset alias in the task spec.

        An alias stands for either an existing dataset or for the
        spec of a future dataset.

        Dataset aliases can be useful in specifying input and output arguments
        to a task.

        Args:
            alias (str): the dataset alias name
            dset  (Dataset|UUID|str|None): specify an existing dataset for this alias
            dspec (dict): the spec for a future dictionary.
        """

        if dset is not None:
            if dspec:
                raise ValueError("Cannot provide both a dataset and a new dataset spec")

            if isinstance(dset, Dataset):
                u = str(dset.id)
            elif isinstance(dset, str):
                u = dset
            elif isinstance(dset, UUID):
                u = str(dset)
            else:
                raise TypeError("Unknown type to specify a Dataset")

            self.datasets[alias] = u

        else:
            if "name" not in dspec:
                raise ValueError("A dataset spec must provide a name")
            if "owner_org" not in dspec:
                dspec["owner_org"] = "stelar-klms"
            if isinstance(dspec["owner_org"], Organization):
                dspec["owner_org"] = str(dspec["owner_org"].id)
            elif isinstance(dspec["owner_org"], UUID):
                dspec["owner_org"] = str(dspec["owner_org"])
            elif not isinstance(dspec["owner_org"], str):
                raise TypeError("Bad type for organization in dataset specification")
            self.datasets[alias] = dspec

        return self

    def process_inspec_entry(
        self, inspec: str | UUID | Resource | list[str | UUID | Resource]
    ) -> Generator[str]:
        if isinstance(inspec, Resource):
            yield str(inspec.id)
        elif isinstance(inspec, str):
            yield inspec
        elif isinstance(inspec, UUID):
            yield str(inspec)
        elif isinstance(inspec, list | tuple):
            for i in inspec:
                yield from self.process_inspec_entry(i)
        else:
            raise TypeError("Unknown type for input spec")

    def process_inspec(self, inspec) -> list[str]:
        return list(self.process_inspec_entry(inspec))

    def i(self, **input_specs):
        for name, inspec in input_specs.items():
            self.inputs[name] = self.process_inspec(inspec)
        return self

    def process_outspec(self, outspec: str | dict):
        # Output is either an S3 URI (str)
        # or a dict. No processing here (yet!)
        if isinstance(outspec, str):
            return {"url": outspec}
        elif isinstance(outspec, dict):
            match outspec:
                case {
                    "url": url,
                    "resource": dict(),
                    "dataset": str(),
                }:
                    return outspec
                case {"url": url, "resource": Resource() as res}:
                    return {"url": url, "resource": str(res.id)}
                case {"url": url, "resource": UUID() as res}:
                    return {"url": url, "resource": str(res)}
                case {"url": url, "resource": str() as res}:
                    return {"url": url, "resource": res}
                case {"url": url}:
                    return outspec
                case _:
                    raise ValueError("Bad format for output spec")
        else:
            raise TypeError("Bad type for output spec, expected string or dict")

    def o(self, **ospec) -> TaskSpec:
        for name, osp in ospec.items():
            self.outputs[name] = self.process_outspec(osp)
        return self

    def p(self, **params) -> TaskSpec:
        self.parameters |= params
        return self

    def spec(self) -> dict:
        s = {}

        if self.tool is not None:
            s["tool"] = self.tool

        if self.image is not None:
            s["image"] = self.image

        s["datasets"] = self.datasets.copy()
        s["inputs"] = self.inputs.copy()
        s["outputs"] = self.outputs.copy()
        s["parameters"] = self.parameters.copy()
        s["name"] = self.name
        return s


class TaskCursor(GenericCursor):
    """A cursor for a collection of STELAR tasks."""

    def __init__(self, api):
        super().__init__(api, Task)

    def fetch(self, **kwargs):
        raise NotImplementedError("TaskCursor does not support fetch operations.")

    def fetch_list(self, **kwargs):
        raise NotImplementedError("TaskCursor does not support fetch operations.")

    def create(self, process_id: UUID | str, task_spec: TaskSpec, secrets=None) -> Task:
        """Create a new Task using the provided TaskSpec.

        Args:
            process_id (UUID): The ID of the process to which the task belongs.
            task_spec (TaskSpec): The specification for the task to create.
            secrets (dict, optional): Secrets to pass to the task.
        Returns:
            Task: The created task.
        """
        json_spec = task_spec.spec()
        json_spec["process_id"] = str(process_id)
        if secrets is not None:
            json_spec["secrets"] = secrets
        ac = api_call(self)
        resp = ac.task_create(**json_spec)
        task_id = resp["id"]
        if resp["job_id"] == "__external__":
            # This is a local task, so we need to handle it differently
            signature = resp["signature"]
            client_for(self).commit_local_task_sig(task_id, signature)
        return self.fetch_proxy(UUID(task_id))
